from series.client.errors import SeriesClientException
from series.logging import Logging

from golgi import configurable

from amino.task import TaskException


@configurable(client=['cli_cmd', 'cli_params'])
class HTTPCLI(Logging):

    def dispatch_command(self, client, cmd_name, params):
        try:
            command = getattr(client, cmd_name)
        except AttributeError:
            self.log.error('No such command: {}'.format(cmd_name))
        else:
            try:
                return command(*params)
            except TypeError as e:
                self.log.error(e)

    def run(self):
        cmd_name = self._cli_cmd[0]
        client = self._client
        try:
            result = self.dispatch_command(client, cmd_name, self._cli_params)
        except SeriesClientException as e:
            self.log.error(e)
        else:
            def err(e):
                if isinstance(e, TaskException):
                    self.log.debug(e)
                else:
                    self.log.error(e)
                return False
            def succ(m):
                if isinstance(m, str):
                    self.log.info(m)
                else:
                    for l in m:
                        self.log.info(l)
                return True
            return result.cata(err, succ)

    @property
    def _client(self):
        pass

__all__ = ['HTTPCLI']



# coding: utf-8

from golgi import configurable
from tek.errors import ParseError

from series.library.handler import BaseHandler
from series.library.model.season import Season
from series.etvdb import ETVDBFacade


@configurable(library=['metadata_interval'])
class Metadata(BaseHandler):

    def __init__(self, library, player):
        super().__init__(library, self._metadata_interval,
                                       'metadata service')
        self._library = library
        self._player = player
        self._min_failed = 0
        self._etvdb = ETVDBFacade()

    def _handle(self, item):
        data = {}
        handler = (self._handle_season if isinstance(item, Season) else
                   self._handle_episode)
        try:
            self.log.info('Fetching metadata for {}…'.format(item))
            handler(item)
        except ParseError as e:
            self.log.error(e)
            data = dict(metadata_failures=item.metadata_failures + 1)
        else:
            data = dict(metadata_fetched=True)
        self._library.alter_object(item, data)

    def _handle_season(self, season):
        sid = self._etvdb.id_by_name(season.series.name)
        if sid:
            data = self._etvdb.season(sid, season.number)
            for episode in data:
                self._update_episode(season.series, episode)

    def _handle_episode(self, episode):
        sid = self._etvdb.id_by_name(episode.series.name)
        if sid:
            data = self._etvdb.episode(episode.series.name, sid,
                                       episode.season.number, episode.number)
            self._update_episode(episode.series, data)

    def _update_episode(self, series, data):
        self._library.alter_episode(series, data['season'], data['episode'],
                                    dict(title=data['title'],
                                         overview=data['overview'],
                                         metadata_fetched=True))

    def _qualify(self, item):
        fails = item.metadata_failures
        return fails is not None and fails == self._min_failed and fails < 15

    @property
    def _candidates(self):
        candidates = (self._seasons or self._new_episodes or
                      self._failed_episodes)
        counts = [c.metadata_failures for c in candidates
                  if c.metadata_failures is not None]
        self._min_failed = min(counts or [0])
        return candidates

    @property
    def _new_episodes(self):
        return [e for e in self._library.new_episodes if not
                e.metadata_fetched]

    @property
    def _seasons(self):
        return self._library.seasons(extra=dict(metadata_fetched=False))

    @property
    def _failed_episodes(self):
        return []


__all__ = ['Metadata']



import socket
from typing import Union
from multiprocessing import Process, Value, Array  # type: ignore

import requests

from golgi.config import configurable
from tek.errors import NotEnoughDiskSpace
from tek.tools import sizeof_fmt

from tek_utils.sharehoster import DownloaderFactory
from tek_utils.sharehoster.errors import ShareHosterError

from amino import List, Empty, Just, Path

from series.get.handler import R, Li, ProcessHandler
from series.condition import DynOr, HandlerCondition
from series.get.model.link import Link, Torrent
from series.get.model.release import ReleaseMonitor
from series.logging import Logging
from series.util import now_unix


class ValidLink(HandlerCondition):

    def __init__(self, link: Union[Torrent, Link]) -> None:
        self.link = link

    @property
    def cond(self):
        if isinstance(self.link, Torrent):
            return self._cond_torrent
        else:
            return self._cond_http

    @property
    def _cond_torrent(self):
        return (~(Li('url', None) | Li('dead') | Li('download_url', None)) &
                Li('cached'))

    @property
    def _cond_http(self):
        return Li('valid')

    def ev(self, item):
        return self.cond.ev(self.link)

    def describe(self, item, target):
        d = self.cond.describe(self.link, target)
        m = self.link.name
        return '{} -> {}'.format(m, d)


class ValidLinks(DynOr):

    @property
    def _sub_type(self):
        return ValidLink

    def _dyn_subs(self, item: ReleaseMonitor):
        return List.wrap(item.all_links)

    def _multiline(self, sub) -> bool:
        return True

    def describe(self, item, target):
        d = super().describe(item, target)
        return 'links => {}'.format(d)


class DownloaderProc(Process, Logging):

    def __init__(self, link, dir):
        self.link = link
        self.dir = dir
        self._downloader_fact = DownloaderFactory()
        self.done = Value('i', 0)
        self.fail = Value('i', 0)
        self.size = Value('d', -1)
        self.file_size = Value('d', -1)
        self.file_path = Array('c', 1024)
        self.error = Array('c', 1024)
        super().__init__(name='seriesd downloader')

    def run(self):
        self._download()
        self.done.value = 1

    def _download(self):
        try:
            dl = self._downloader_fact(self.link.download_url,
                                       download_dir=str(self.dir))
            dl.retrieve()
            self.size.value = dl.outfile_size
            self.file_size.value = dl.file_size
            self.file_path.value = str(dl.file_path).encode()
        except (NotEnoughDiskSpace, requests.RequestException,
                ShareHosterError, ConnectionError, socket.timeout) as e:
            self.error.value = str(e).encode()
            self.fail.value = 1


@configurable(get=['download_dir', 'min_size', 'download_timeout'])
class Downloader(ProcessHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(self._download_timeout, releases, 5,
                         'download manager')
        self._min_failed = 0
        self._proc = Empty()

    def _fail(self, monitor, link, err):
        self.log.error('Downloader: {}'.format(err))
        f = monitor.failed_downloads + 1
        self._update(monitor, failed_downloads=f)
        self._releases.fail_link(link.id)

    def _create_async(self, monitor):
        self.log.info('Downloading {!s}.'.format(monitor.release))
        self._update(monitor, downloading=True)
        link = monitor.link
        self._download_dir.mkdir(parents=True, exist_ok=True)
        return Just(DownloaderProc(link, self._download_dir))

    def _clean_done(self, a):
        proc = a.proc
        if proc.done.value:
            self._update_id(a.monitor.id, downloading=False)
            if proc.fail.value:
                self._fail(a.monitor, proc.link, proc.error.value)
            else:
                path = Path(proc.file_path.value.decode())
                return self._check_download(a.monitor, proc.link,
                                            proc.size.value,
                                            proc.file_size.value, path)

    def _clean_timeout(self, a):
        self._update_id(a.monitor.id, downloading=False)
        self._fail(a.monitor, a.proc.link, 'timeout exceeded')

    def _check_download(self, monitor, link, size, file_size, file_path):
        ''' Verify the downloaded file's size by means of two criteria:
        The content-length header sent from the hoster must match
        exactly.
        The config parameter min_size defines what is considered to be a
        faulty file, i.e. the link checker didn't detect the file to be
        invalid, but it is small enough to just be the hoster's error
        page.
        '''
        if size < self._min_size:
            msg = 'File smaller than required ({}): {!s}'
            self._fail(monitor, link,
                       msg.format(sizeof_fmt(size), monitor.release))
        elif size != file_size and file_size != -1:
            msg = 'Filesize mismatch in download of {!s}:'
            self._fail(monitor, link, msg.format(monitor.release))
            msg = 'Got {}, expected {}.'
            self.log.error(msg.format(size, file_size))
            try:
                file_path.unlink()
            except IOError:
                msg = 'Couldn\'t delete broken download of {}.'
                self.log.warn(msg.format(monitor.release))
        else:
            self._releases.update_by_id(monitor.id,
                                        download_path=str(file_path),
                                        download_date_stamp=now_unix())
            self._releases.mark_episode_downloaded(monitor)
            return True

    @property
    def _conditions(self):
        return (~(R('downloaded') | R('nuked') | R('downloading')) &
                ValidLinks())

    def _choose(self, candidates):
        counts = [r.item.failed_downloads for r in candidates]
        min_failed = min(counts or [0])
        best = candidates.filter(
            lambda a: a.item.failed_downloads == min_failed)
        return best.head

    @property
    def _candidates(self):
        return self._releases.monitors.filter_by(downloaded=False).all()

__all__ = ('Downloader',)



import abc
from datetime import datetime, timedelta

from series.condition import AttrCondition
from series.handler import Handler
from series.get.model.release import ReleaseMonitor
from series.get.model.link import Link, Torrent
from series.get.model.show import Show

from sqlalchemy.sql.expression import and_

import amino
from amino import Empty, Maybe, _, __


class ReleaseAttr(AttrCondition[ReleaseMonitor]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('release', attr, target=target)


R = ReleaseAttr


class LinkAttr(AttrCondition[Link]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('link', attr, target)


Li = LinkAttr


class ShowAttr(AttrCondition[Show]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('show', attr, target)


S = ShowAttr


class BaseHandler(Handler):

    def __init__(self, data, interval, description, **kw):
        self._data = data
        super().__init__(interval, description, **kw)

    @property
    def _candidates(self):
        return self._data.all

    @property
    def _lock(self):
        return self._data.lock

    def _update(self, item, **data):
        self._update_id(item.id, **data)

    def _update_id(self, id, **data):
        self._data.update_by_id(id, **data)


class ReleaseHandler(BaseHandler):

    @property
    def _releases(self):
        return self._data

    @property
    def _no_cached_torrents_q(self):
        return (
            self._releases.monitors
            .filter_by(downloaded=False, downloading=False)
            .filter(~ReleaseMonitor.torrents.any(
                and_(Torrent.cached, ~Torrent.dead)))
        )

    @property
    def _no_cached_torrents(self):
        return self._no_cached_torrents_q.all()


class ShowHandler(BaseHandler):

    @property
    def _shows(self):
        return self._data


class AsyncAdapter:

    def __init__(self, proc, monitor) -> None:
        self.proc = proc
        self.monitor = monitor
        self.created = datetime.now()

    def kill(self):
        pass

    def start(self):
        return self.proc.start()


class ProcessAdapter(AsyncAdapter):

    def kill(self):
        return self.proc.terminate()


class AsyncHandler(ReleaseHandler):

    def __init__(self, timeout, *a, **kw) -> None:
        super().__init__(*a, **kw)
        self._async = Empty()
        self._timeout = timedelta(minutes=timeout)

    def _handle(self, monitor):
        if self._async.empty:
            self._async = (
                self._create_async(monitor) /
                amino.L(self._adapter)(_, monitor) %
                __.start()
            )

    def _cleanup(self):
        def timeout(a):
            if datetime.now() - a.created > self._timeout:
                self._clean_timeout(a)
                a.kill()
                return True
        if (self._async.exists(self._clean_done) or
                self._async.exists(timeout)):
            self._async = Empty()

    @abc.abstractproperty
    def _adapter(self) -> type:
        ...

    @abc.abstractmethod
    def _clean_done(self, proc) -> bool:
        ...

    @abc.abstractmethod
    def _clean_timeout(self, proc) -> None:
        ...


class ThreadHandler(AsyncHandler):

    @property
    def _adapter(self):
        return AsyncAdapter


class ProcessHandler(AsyncHandler):

    @property
    def _adapter(self):
        return ProcessAdapter

    @abc.abstractmethod
    def _create_async(self, monitor) -> Maybe[AsyncAdapter]:
        ...


__all__ = ('ReleaseHandler', 'ShowHandler', 'ProcessHandler', 'ThreadHandler')
